#define ARMA_NO_DEBUG
#define ARMA_DONT_USE_OPENMP
#define STRICT_R_HEADERS // needed on Windows, not macOS
#include <RcppArmadillo.h>
#include <omp.h>
#include <RcppArmadilloExtensions/sample.h> // for Rcpp::RcppArmadillo::sample
// [[Rcpp::depends(RcppArmadillo)]]
// [[Rcpp::plugins(cpp11)]]
// [[Rcpp::plugins(openmp)]]

using namespace Rcpp;
using namespace arma;
using namespace std;

// [[Rcpp::export]]
double logplkd_fixtra_standalone(const vec &event, const IntegerVector &count_strata,
                                 const mat &Z_tv, const mat &B_spline, const mat &theta,
                                 const mat &Z_ti, const vec &beta_ti,
                                 const bool &parallel=false, const unsigned int &threads=1) {
  bool ti = norm(Z_ti, "inf") > sqrt(datum::eps);
  IntegerVector cumsum_strata = cumsum(count_strata);
  unsigned int n_strata = cumsum_strata.length();
  cumsum_strata.push_front(0);
  double logplkd = 0.0;
  for (unsigned int i = 0; i < n_strata; ++i) {
    unsigned int start = cumsum_strata[i], end = cumsum_strata[i+1];
    uvec idx_fail = find(event.rows(start, end-1)==1);
    mat Z_tv_theta = Z_tv.rows(start+idx_fail(0), end-1) * theta;
    vec Z_ti_beta_ti;
    if (ti) {
      Z_ti_beta_ti = Z_ti.rows(start+idx_fail(0), end-1) * beta_ti;
    }
    mat B_sp = B_spline.rows(start+idx_fail);
    idx_fail -= idx_fail(0); // index within the strata
    unsigned int n_fail = idx_fail.n_elem, n_Z_tv_theta = Z_tv_theta.n_rows;
    vec lincomb_fail(n_fail);
    if (parallel) {
      double scale_fac = as_scalar(idx_fail.tail(1));
      vec cumsum_ar = (double)n_Z_tv_theta / scale_fac * regspace(1,n_fail) -
        cumsum(conv_to<vec>::from(idx_fail)/scale_fac); // cum sum of at risk cts
      omp_set_num_threads(threads);
      #pragma omp parallel
      {
        unsigned int id, j, istart, iend;
        id = omp_get_thread_num();
        istart = as_scalar(
          find(cumsum_ar >= cumsum_ar(n_fail-1)/(double)threads*id, 1));
        // exclusive
        iend = as_scalar(
          find(cumsum_ar >= cumsum_ar(n_fail-1)/(double)threads*(id+1), 1));
        if (id==threads-1) iend = n_fail;
        double val_tmp = 0;
        if (ti) {
          for (j = istart; j < iend; ++j) {
            vec lincomb =
              Z_tv_theta.rows(idx_fail(j), n_Z_tv_theta-1)*B_sp.row(j).t() +
              Z_ti_beta_ti.subvec(idx_fail(j), n_Z_tv_theta-1);
            lincomb_fail(j) = lincomb(0);
            val_tmp += log(sum(exp(lincomb)));
          }
        } else {
          for (j = istart; j < iend; ++j) {
            vec lincomb =
              Z_tv_theta.rows(idx_fail(j), n_Z_tv_theta-1) * B_sp.row(j).t();
            lincomb_fail(j) = lincomb(0);
            val_tmp += log(sum(exp(lincomb)));
          }
        }
        #pragma omp atomic
        logplkd -= val_tmp;
      }
    } else {
      if (ti) {
        for (unsigned int j = 0; j < n_fail; ++j) {
          vec lincomb =
            Z_tv_theta.rows(idx_fail(j), n_Z_tv_theta-1) * B_sp.row(j).t() +
            Z_ti_beta_ti.subvec(idx_fail(j), n_Z_tv_theta-1);
          lincomb_fail(j) = lincomb(0);
          logplkd -= log(sum(exp(lincomb)));
        }
      } else {
        for (unsigned int j = 0; j < n_fail; ++j) {
          vec lincomb =
            Z_tv_theta.rows(idx_fail(j), n_Z_tv_theta-1) * B_sp.row(j).t();
          lincomb_fail(j) = lincomb(0);
          logplkd -= log(sum(exp(lincomb)));
        }
      }
    }
    logplkd += accu(lincomb_fail);
  }
  logplkd /= Z_tv.n_rows;
  return logplkd;
}

// [[Rcpp::export]]
double logplkd_sim(const vec &theta_vec, const vec &event, 
                   const IntegerVector &count_strata,
                   const mat &Z_tv, const mat &B_spline,
                   const mat &Z_ti, const vec &beta_ti,
                   const bool &parallel=false, const unsigned int &threads=1) {
  // data should be ordered by strata first and then time
  // Z_tv: design matrix corresponding to time-varying coefs
  // B_spline: a matrix of B-splines, # of time points as row ct,
  //           and # of knots as col ct
  // theta: p by k; B_spline: t by k
  // Z_ti: design matrix corresponding to time-invariant coefs
  // beta_ti: a vector of initial values of time-invariant coefs
  mat theta = reshape(theta_vec, Z_tv.n_cols, B_spline.n_cols);
  bool ti = norm(Z_ti, "inf") > sqrt(datum::eps);
  IntegerVector cumsum_strata = cumsum(count_strata);
  unsigned int n_strata = cumsum_strata.length();
  cumsum_strata.push_front(0);
  double logplkd = 0.0;
  for (unsigned int i = 0; i < n_strata; ++i) {
    unsigned int start = cumsum_strata[i], end = cumsum_strata[i+1];
    uvec idx_fail = find(event.rows(start, end-1)==1);
    mat Z_tv_theta = Z_tv.rows(start+idx_fail(0), end-1) * theta;
    vec Z_ti_beta_ti;
    if (ti) {
      Z_ti_beta_ti = Z_ti.rows(start+idx_fail(0), end-1) * beta_ti;
    }
    mat B_sp = B_spline.rows(start+idx_fail);
    idx_fail -= idx_fail(0); // index within the strata
    unsigned int n_fail = idx_fail.n_elem, n_Z_tv_theta = Z_tv_theta.n_rows;
    vec lincomb_fail(n_fail);
    if (parallel) {
      double scale_fac = as_scalar(idx_fail.tail(1));
      vec cumsum_ar = (double)n_Z_tv_theta / scale_fac * regspace(1,n_fail) -
        cumsum(conv_to<vec>::from(idx_fail)/scale_fac); // cum sum of at risk cts
      omp_set_num_threads(threads);
      #pragma omp parallel
      {
        unsigned int id, j, istart, iend;
        id = omp_get_thread_num();
        istart = as_scalar(
          find(cumsum_ar >= cumsum_ar(n_fail-1)/(double)threads*id, 1));
        // exclusive
        iend = as_scalar(
          find(cumsum_ar >= cumsum_ar(n_fail-1)/(double)threads*(id+1), 1));
        if (id==threads-1) iend = n_fail;
        double val_tmp = 0;
        if (ti) {
          for (j = istart; j < iend; ++j) {
            vec lincomb =
              Z_tv_theta.rows(idx_fail(j), n_Z_tv_theta-1)*B_sp.row(j).t() +
              Z_ti_beta_ti.subvec(idx_fail(j), n_Z_tv_theta-1);
            lincomb_fail(j) = lincomb(0);
            val_tmp += log(sum(exp(lincomb)));
          }
        } else {
          for (j = istart; j < iend; ++j) {
            vec lincomb =
              Z_tv_theta.rows(idx_fail(j), n_Z_tv_theta-1) * B_sp.row(j).t();
            lincomb_fail(j) = lincomb(0);
            val_tmp += log(sum(exp(lincomb)));
          }
        }
        #pragma omp atomic
        logplkd -= val_tmp;
      }
    } else {
      if (ti) {
        for (unsigned int j = 0; j < n_fail; ++j) {
          vec lincomb =
            Z_tv_theta.rows(idx_fail(j), n_Z_tv_theta-1) * B_sp.row(j).t() +
            Z_ti_beta_ti.subvec(idx_fail(j), n_Z_tv_theta-1);
          lincomb_fail(j) = lincomb(0);
          logplkd -= log(sum(exp(lincomb)));
        }
      } else {
        for (unsigned int j = 0; j < n_fail; ++j) {
          vec lincomb =
            Z_tv_theta.rows(idx_fail(j), n_Z_tv_theta-1) * B_sp.row(j).t();
          lincomb_fail(j) = lincomb(0);
          logplkd -= log(sum(exp(lincomb)));
        }
      }
    }
    logplkd += accu(lincomb_fail);
  }
  logplkd /= Z_tv.n_rows;
  return logplkd;
}

// [[Rcpp::export]]
List gradinfo_fixtra_standalone(const vec &event, const IntegerVector &count_strata,
                                const mat &Z_tv, const mat &B_spline, const mat &theta,
                                const mat &Z_ti, const vec &beta_ti,
                                const string &method="Newton", const double &lambda=1e8,
                                const bool &parallel=false, const unsigned int &threads=1) {
  // data should be ordered by strata first and then time
  // theta: p by k; B_spline: t by k
  
  bool ti = norm(Z_ti, "inf") > sqrt(datum::eps);
  IntegerVector cumsum_strata = cumsum(count_strata);
  unsigned int n_strata = cumsum_strata.length();
  cumsum_strata.push_front(0);
  vec grad; mat info, gradd(theta.n_elem, int(accu(event)),fill::zeros);
  if (ti) {
    grad = zeros<vec>(theta.n_elem+beta_ti.n_elem);
    info = zeros<mat>(theta.n_elem+beta_ti.n_elem, theta.n_elem+beta_ti.n_elem);
  } else {
    grad = zeros<vec>(theta.n_elem);
    info = zeros<mat>(theta.n_elem, theta.n_elem);
  }
  for (unsigned int i = 0; i < n_strata; ++i) {
    uvec idx_fail =
      find(event.rows(cumsum_strata[i], cumsum_strata[i+1]-1)==1);
    mat Z_tv_strata =
      Z_tv.rows(cumsum_strata[i]+idx_fail(0), cumsum_strata[i+1]-1);
    mat Z_ti_strata; vec Z_ti_beta_ti;
    if (ti) {
      Z_ti_strata =
        Z_ti.rows(cumsum_strata[i]+idx_fail(0), cumsum_strata[i+1]-1);
      Z_ti_beta_ti = Z_ti_strata * beta_ti;
    }
    mat B_sp = B_spline.rows(cumsum_strata[i]+idx_fail);
    mat Z_tv_theta = Z_tv_strata * theta;
    idx_fail -= idx_fail(0); // index within the strata
    unsigned int n_fail = idx_fail.n_elem, n_Z_tv_theta = Z_tv_theta.n_rows;
    if (parallel) {
      double scale_fac = as_scalar(idx_fail.tail(1));
      vec cumsum_ar = (double)n_Z_tv_theta / scale_fac * regspace(1,n_fail) -
        cumsum(conv_to<vec>::from(idx_fail)/scale_fac); // cum sum of at risk counts
      omp_set_num_threads(threads);
      #pragma omp parallel
      {
        unsigned int id, j, istart, iend;
        id = omp_get_thread_num();
        istart =
          as_scalar(find(cumsum_ar>=cumsum_ar(n_fail-1)/(double)threads*id, 1));
        iend =
          as_scalar(find(cumsum_ar>=cumsum_ar(n_fail-1)/(double)threads*(id+1), 1)); // exclusive
        if (id == threads-1) iend = n_fail;
        vec grad_tmp(size(grad), fill::zeros);
        mat info_tmp(size(info), fill::zeros);
        if (ti) {
          for (j = istart; j < iend; ++j) {
            vec B_sp_tmp = B_sp.row(j).t();
            unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
            vec exp_lincomb = exp(Z_tv_theta.rows(arstart,arend) * B_sp_tmp +
                                    Z_ti_beta_ti.subvec(arstart,arend));
            double S0 = accu(exp_lincomb);
            mat Z_tv_exp =
              Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
            mat Z_ti_exp =
              Z_ti_strata.rows(arstart,arend).each_col() % exp_lincomb;
            vec S1_tv = sum(Z_tv_exp).t(), S1_ti = sum(Z_ti_exp).t();
            vec S1 = join_cols(kron(S1_tv,B_sp_tmp), S1_ti)/S0;
            mat S2_11 =
              kron(Z_tv_strata.rows(arstart,arend).t()*Z_tv_exp,
                   B_sp_tmp*B_sp_tmp.t())/S0,
            S2_12 =
              kron(Z_tv_strata.rows(arstart,arend).t()*Z_ti_exp,
                   B_sp_tmp)/S0,
            S2_22 = Z_ti_strata.rows(arstart,arend).t()*Z_ti_exp/S0;
            grad_tmp += join_cols(kron(Z_tv_strata.row(arstart).t(), B_sp_tmp),
                                  Z_ti_strata.row(arstart).t()) - S1;
            info_tmp += join_cols(join_rows(S2_11, S2_12),
                                  join_rows(S2_12.t(), S2_22)) - S1*S1.t();
          }
        } else {
          for (j = istart; j < iend; ++j) {
            vec B_sp_tmp = B_sp.row(j).t();
            unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
            vec exp_lincomb =
              exp(Z_tv_theta.rows(arstart,arend) * B_sp_tmp);
            double S0 = accu(exp_lincomb);
            mat Z_tv_exp =
              Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
            vec S1_tv = sum(Z_tv_exp).t();
            mat S2 = Z_tv_strata.rows(arstart,arend).t() * Z_tv_exp;
            grad_tmp += kron(Z_tv_strata.row(arstart).t()-S1_tv/S0, B_sp_tmp);
            info_tmp +=
              kron(S2/S0-S1_tv*S1_tv.t()/pow(S0,2),B_sp_tmp*B_sp_tmp.t());
          }
        }
        #pragma omp critical (gradient)
        grad += grad_tmp;
        #pragma omp critical (information)
        info += info_tmp;
      }
    } else {
      if (ti) {
        for (unsigned int j = 0; j < n_fail; ++j) {
          vec B_sp_tmp = B_sp.row(j).t();
          unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
          vec exp_lincomb = exp(Z_tv_theta.rows(arstart,arend) * B_sp_tmp +
                                  Z_ti_beta_ti.subvec(arstart,arend));
          double S0 = accu(exp_lincomb);
          mat Z_tv_exp =
            Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
          mat Z_ti_exp =
            Z_ti_strata.rows(arstart,arend).each_col() % exp_lincomb;
          vec S1_tv = sum(Z_tv_exp).t(), S1_ti = sum(Z_ti_exp).t();
          vec S1 = join_cols(kron(S1_tv,B_sp_tmp), S1_ti)/S0;
          mat S2_11 =
            kron(Z_tv_strata.rows(arstart,arend).t()*Z_tv_exp,
                 B_sp_tmp*B_sp_tmp.t())/S0,
          S2_12 =
            kron(Z_tv_strata.rows(arstart,arend).t()*Z_ti_exp,
                 B_sp_tmp)/S0,
          S2_22 = Z_ti_strata.rows(arstart,arend).t()*Z_ti_exp/S0;
          grad += join_cols(kron(Z_tv_strata.row(arstart).t(), B_sp_tmp),
                            Z_ti_strata.row(arstart).t()) - S1;
          info += join_cols(join_rows(S2_11, S2_12),
                            join_rows(S2_12.t(), S2_22)) - S1*S1.t();
        }
      } else {
        for (unsigned int j = 0; j < n_fail; ++j) {
          vec B_sp_tmp = B_sp.row(j).t();
          unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
          vec exp_lincomb =
            exp(Z_tv_theta.rows(arstart, arend) * B_sp_tmp);
          double S0 = accu(exp_lincomb);
          mat Z_tv_exp =
            Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
          vec S1_tv = sum(Z_tv_exp).t();
          mat S2 = Z_tv_strata.rows(arstart,arend).t() * Z_tv_exp;
          grad += kron(Z_tv_strata.row(arstart).t() - S1_tv/S0, B_sp_tmp);
          gradd.col(j) += kron(Z_tv_strata.row(arstart).t() - S1_tv/S0, B_sp_tmp);
          info += kron(S2/S0-S1_tv*S1_tv.t()/pow(S0,2),B_sp_tmp*B_sp_tmp.t());
        }
      }
    }
  }
  
  vec step; // Newton step
  if (method=="ProxN") {
    info.diag() += Z_tv.n_rows / lambda;
    step = solve(info, grad, solve_opts::fast+solve_opts::likely_sympd);
    info.diag() -= Z_tv.n_rows / lambda;
  } else if (method=="Newton") {
    step = solve(info, grad, solve_opts::fast+solve_opts::likely_sympd);
  }
  double inc = dot(grad, step) / Z_tv.n_rows; // square of Newton increment
  if (ti) {
    return List::create(_["step_tv"]=reshape(step.head(theta.n_elem),
                                             size(theta.t())).t(),
                         _["step_ti"]=step.tail(beta_ti.n_elem),
                         _["grad"]=grad, _["info"]=info,
                         _["inc"]=inc);
  } else {
    return List::create(_["step"]=reshape(step, size(theta.t())).t(),
                         _["grad"]=grad, _["gradd"]=gradd, _["info"]=info, _["inc"]=inc);
  }
}

// [[Rcpp::export]]
mat info_sim(const mat &theta, const vec &event, 
             const IntegerVector &count_strata,
             const mat &Z_tv, const mat &B_spline, 
             const mat &Z_ti, const vec &beta_ti,
             const bool &parallel=true, const unsigned int &threads=1) {
  // data should be ordered by strata first and then time
  // theta: p by k; B_spline: t by k
  
  bool ti = norm(Z_ti, "inf") > sqrt(datum::eps);
  IntegerVector cumsum_strata = cumsum(count_strata);
  unsigned int n_strata = cumsum_strata.length();
  cumsum_strata.push_front(0);
  mat info;
  if (ti) {
    info = zeros<mat>(theta.n_elem+beta_ti.n_elem, theta.n_elem+beta_ti.n_elem);
  } else {
    info = zeros<mat>(theta.n_elem, theta.n_elem);
  }
  for (unsigned int i = 0; i < n_strata; ++i) {
    uvec idx_fail =
      find(event.rows(cumsum_strata[i], cumsum_strata[i+1]-1)==1);
    mat Z_tv_strata =
      Z_tv.rows(cumsum_strata[i]+idx_fail(0), cumsum_strata[i+1]-1);
    mat Z_ti_strata; vec Z_ti_beta_ti;
    if (ti) {
      Z_ti_strata =
        Z_ti.rows(cumsum_strata[i]+idx_fail(0), cumsum_strata[i+1]-1);
      Z_ti_beta_ti = Z_ti_strata * beta_ti;
    }
    mat B_sp = B_spline.rows(cumsum_strata[i]+idx_fail);
    mat Z_tv_theta = Z_tv_strata * theta;
    idx_fail -= idx_fail(0); // index within the strata
    unsigned int n_fail = idx_fail.n_elem, n_Z_tv_theta = Z_tv_theta.n_rows;
    if (parallel) {
      double scale_fac = as_scalar(idx_fail.tail(1));
      vec cumsum_ar = (double)n_Z_tv_theta / scale_fac * regspace(1,n_fail) -
        cumsum(conv_to<vec>::from(idx_fail)/scale_fac); // cum sum of at risk counts
      omp_set_num_threads(threads);
      #pragma omp parallel
      {
        unsigned int id, j, istart, iend;
        id = omp_get_thread_num();
        istart =
          as_scalar(find(cumsum_ar>=cumsum_ar(n_fail-1)/(double)threads*id, 1));
        iend =
          as_scalar(find(cumsum_ar>=cumsum_ar(n_fail-1)/(double)threads*(id+1), 1)); // exclusive
        if (id == threads-1) iend = n_fail;
        mat info_tmp(size(info), fill::zeros);
        if (ti) {
          for (j = istart; j < iend; ++j) {
            vec B_sp_tmp = B_sp.row(j).t();
            unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
            vec exp_lincomb = exp(Z_tv_theta.rows(arstart,arend) * B_sp_tmp +
                                    Z_ti_beta_ti.subvec(arstart,arend));
            double S0 = accu(exp_lincomb);
            mat Z_tv_exp =
              Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
            mat Z_ti_exp =
              Z_ti_strata.rows(arstart,arend).each_col() % exp_lincomb;
            vec S1_tv = sum(Z_tv_exp).t(), S1_ti = sum(Z_ti_exp).t();
            vec S1 = join_cols(kron(S1_tv,B_sp_tmp), S1_ti)/S0;
            mat S2_11 =
              kron(Z_tv_strata.rows(arstart,arend).t()*Z_tv_exp,
                   B_sp_tmp*B_sp_tmp.t())/S0,
            S2_12 =
              kron(Z_tv_strata.rows(arstart,arend).t()*Z_ti_exp,
                   B_sp_tmp)/S0,
            S2_22 = Z_ti_strata.rows(arstart,arend).t()*Z_ti_exp/S0;
            info_tmp += join_cols(join_rows(S2_11, S2_12),
                                  join_rows(S2_12.t(), S2_22)) - S1*S1.t();
          }
        } else {
          for (j = istart; j < iend; ++j) {
            vec B_sp_tmp = B_sp.row(j).t();
            unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
            vec exp_lincomb =
              exp(Z_tv_theta.rows(arstart,arend) * B_sp_tmp);
            double S0 = accu(exp_lincomb);
            mat Z_tv_exp =
              Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
            vec S1_tv = sum(Z_tv_exp).t();
            mat S2 = Z_tv_strata.rows(arstart,arend).t() * Z_tv_exp;
            info_tmp +=
              kron(S2/S0-S1_tv*S1_tv.t()/pow(S0,2),B_sp_tmp*B_sp_tmp.t());
          }
        }
        #pragma omp critical (information)
        info += info_tmp;
      }
    } else {
      if (ti) {
        for (unsigned int j = 0; j < n_fail; ++j) {
          vec B_sp_tmp = B_sp.row(j).t();
          unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
          vec exp_lincomb = exp(Z_tv_theta.rows(arstart,arend) * B_sp_tmp +
                                  Z_ti_beta_ti.subvec(arstart,arend));
          double S0 = accu(exp_lincomb);
          mat Z_tv_exp =
            Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
          mat Z_ti_exp =
            Z_ti_strata.rows(arstart,arend).each_col() % exp_lincomb;
          vec S1_tv = sum(Z_tv_exp).t(), S1_ti = sum(Z_ti_exp).t();
          vec S1 = join_cols(kron(S1_tv,B_sp_tmp), S1_ti)/S0;
          mat S2_11 =
            kron(Z_tv_strata.rows(arstart,arend).t()*Z_tv_exp,
                 B_sp_tmp*B_sp_tmp.t())/S0,
          S2_12 =
            kron(Z_tv_strata.rows(arstart,arend).t()*Z_ti_exp,
                 B_sp_tmp)/S0,
          S2_22 = Z_ti_strata.rows(arstart,arend).t()*Z_ti_exp/S0;
          info += join_cols(join_rows(S2_11, S2_12),
                            join_rows(S2_12.t(), S2_22)) - S1*S1.t();
        }
      } else {
        for (unsigned int j = 0; j < n_fail; ++j) {
          vec B_sp_tmp = B_sp.row(j).t();
          unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
          vec exp_lincomb =
            exp(Z_tv_theta.rows(arstart, arend) * B_sp_tmp);
          double S0 = accu(exp_lincomb);
          mat Z_tv_exp =
            Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
          vec S1_tv = sum(Z_tv_exp).t();
          mat S2 = Z_tv_strata.rows(arstart,arend).t() * Z_tv_exp;
          info += kron(S2/S0-S1_tv*S1_tv.t()/pow(S0,2),B_sp_tmp*B_sp_tmp.t());
        }
      }
    }
  }
  info.diag() += sqrt(datum::eps);
  return info;
}

// [[Rcpp::export]]
vec grad_sim(const vec &theta_vec, const vec &event,
             const IntegerVector &count_strata,
             const mat &Z_tv, const mat &B_spline, 
             const mat &Z_ti, const vec &beta_ti,
             const bool &parallel=true, const unsigned int &threads=1) {
  // data should be ordered by strata first and then time
  // theta: p by k; B_spline: t by k
  mat theta = reshape(theta_vec, Z_tv.n_cols, B_spline.n_cols);
  bool ti = norm(Z_ti, "inf") > sqrt(datum::eps);
  IntegerVector cumsum_strata = cumsum(count_strata);
  unsigned int n_strata = cumsum_strata.length();
  cumsum_strata.push_front(0);
  vec grad;
  if (ti) {
    grad = zeros<vec>(theta.n_elem+beta_ti.n_elem);
  } else {
    grad = zeros<vec>(theta.n_elem);
  }
  for (unsigned int i = 0; i < n_strata; ++i) {
    uvec idx_fail =
      find(event.rows(cumsum_strata[i], cumsum_strata[i+1]-1)==1);
    mat Z_tv_strata =
      Z_tv.rows(cumsum_strata[i]+idx_fail(0), cumsum_strata[i+1]-1);
    mat Z_ti_strata; vec Z_ti_beta_ti;
    if (ti) {
      Z_ti_strata =
        Z_ti.rows(cumsum_strata[i]+idx_fail(0), cumsum_strata[i+1]-1);
      Z_ti_beta_ti = Z_ti_strata * beta_ti;
    }
    mat B_sp = B_spline.rows(cumsum_strata[i]+idx_fail);
    mat Z_tv_theta = Z_tv_strata * theta;
    idx_fail -= idx_fail(0); // index within the strata
    unsigned int n_fail = idx_fail.n_elem, n_Z_tv_theta = Z_tv_theta.n_rows;
    if (parallel) {
      double scale_fac = as_scalar(idx_fail.tail(1));
      vec cumsum_ar = (double)n_Z_tv_theta / scale_fac * regspace(1,n_fail) -
        cumsum(conv_to<vec>::from(idx_fail)/scale_fac); // cum sum of at risk counts
      omp_set_num_threads(threads);
      #pragma omp parallel
      {
        unsigned int id, j, istart, iend;
        id = omp_get_thread_num();
        istart =
          as_scalar(find(cumsum_ar>=cumsum_ar(n_fail-1)/(double)threads*id, 1));
        iend =
          as_scalar(find(cumsum_ar>=cumsum_ar(n_fail-1)/(double)threads*(id+1), 1)); // exclusive
        if (id == threads-1) iend = n_fail;
        vec grad_tmp(size(grad), fill::zeros);
        if (ti) {
          for (j = istart; j < iend; ++j) {
            vec B_sp_tmp = B_sp.row(j).t();
            unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
            vec exp_lincomb = exp(Z_tv_theta.rows(arstart,arend) * B_sp_tmp +
                                    Z_ti_beta_ti.subvec(arstart,arend));
            double S0 = accu(exp_lincomb);
            mat Z_tv_exp =
              Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
            mat Z_ti_exp =
              Z_ti_strata.rows(arstart,arend).each_col() % exp_lincomb;
            vec S1_tv = sum(Z_tv_exp).t(), S1_ti = sum(Z_ti_exp).t();
            vec S1 = join_cols(kron(S1_tv,B_sp_tmp), S1_ti)/S0;
            grad_tmp += join_cols(kron(Z_tv_strata.row(arstart).t(), B_sp_tmp),
                                  Z_ti_strata.row(arstart).t()) - S1;
          }
        } else {
          for (j = istart; j < iend; ++j) {
            vec B_sp_tmp = B_sp.row(j).t();
            unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
            vec exp_lincomb =
              exp(Z_tv_theta.rows(arstart,arend) * B_sp_tmp);
            double S0 = accu(exp_lincomb);
            mat Z_tv_exp =
              Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
            vec S1_tv = sum(Z_tv_exp).t();
            grad_tmp += kron(Z_tv_strata.row(arstart).t()-S1_tv/S0, B_sp_tmp);
          }
        }
        #pragma omp critical (gradient)
        grad += grad_tmp;
      }
    } else {
      if (ti) {
        for (unsigned int j = 0; j < n_fail; ++j) {
          vec B_sp_tmp = B_sp.row(j).t();
          unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
          vec exp_lincomb = exp(Z_tv_theta.rows(arstart,arend) * B_sp_tmp +
                                  Z_ti_beta_ti.subvec(arstart,arend));
          double S0 = accu(exp_lincomb);
          mat Z_tv_exp =
            Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
          mat Z_ti_exp =
            Z_ti_strata.rows(arstart,arend).each_col() % exp_lincomb;
          vec S1_tv = sum(Z_tv_exp).t(), S1_ti = sum(Z_ti_exp).t();
          vec S1 = join_cols(kron(S1_tv,B_sp_tmp), S1_ti)/S0;
          grad += join_cols(kron(Z_tv_strata.row(arstart).t(), B_sp_tmp),
                            Z_ti_strata.row(arstart).t()) - S1;
        }
      } else {
        for (unsigned int j = 0; j < n_fail; ++j) {
          vec B_sp_tmp = B_sp.row(j).t();
          unsigned int arstart = idx_fail(j), arend = n_Z_tv_theta-1; // at risk set indices
          vec exp_lincomb =
            exp(Z_tv_theta.rows(arstart, arend) * B_sp_tmp);
          double S0 = accu(exp_lincomb);
          mat Z_tv_exp =
            Z_tv_strata.rows(arstart,arend).each_col() % exp_lincomb;
          vec S1_tv = sum(Z_tv_exp).t();
          grad += kron(Z_tv_strata.row(arstart).t() - S1_tv/S0, B_sp_tmp);
        }
      }
    }
  }
  return grad;
}

List logplkd_fixtra_bresties(const mat &Z_tv, const mat &B1, const mat &B2,
                             const mat &theta,
                             const mat &Z_ti, const vec &beta_ti, const bool &ti,
                             const unsigned int n_strata,
                             vector<uvec> &idx_B_sp, vector<vector<uvec>> &idx_fail,
                             vector<unsigned int> n_Z_strata,
                             vector<vector<unsigned int>> &idx_Z_strata,
                             vector<vector<unsigned int>> &istart,
                             vector<vector<unsigned int>> &iend,
                             const bool &parallel=false, const unsigned int &threads=1) {
  // theta: p by K_1K_2
  // B1: calendar time
  // B2: post-discharge time
  double norm_parm;
  vector<vec> hazard;
  if (ti) {
    norm_parm = max(norm(theta, "inf"), norm(beta_ti, "inf"));
  } else {
    norm_parm = norm(theta, "inf");
  }
  double logplkd = 0.0;
  if (norm_parm < sqrt(datum::eps)) { // theta and beta_ti are 0
    if (parallel) {
      for (unsigned int i = 0; i < n_strata; ++i) {
        vec hazard_tmp(idx_fail[i].size());
        omp_set_num_threads(threads);
        #pragma omp parallel
        {
          unsigned int id, j;
          id = omp_get_thread_num();
          double val_tmp = 0;
          for (j = istart[i][id]; j < iend[i][id]; ++j) {
            double tmp = n_Z_strata[i]-idx_fail[i][j](0);
            hazard_tmp(j) = idx_fail[i][j].n_elem/tmp;
            val_tmp += idx_fail[i][j].n_elem*log(tmp);
          }
          #pragma omp atomic
          logplkd -= val_tmp;
        }
        hazard.push_back(hazard_tmp);
      }
    } else {
      for (unsigned int i = 0; i < n_strata; ++i) {
        vec hazard_tmp(idx_fail[i].size());
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          double tmp = n_Z_strata[i]-idx_fail[i][j](0);
          hazard_tmp(j) = idx_fail[i][j].n_elem/tmp;
          logplkd -= idx_fail[i][j].n_elem*log(tmp);
        }
        hazard.push_back(hazard_tmp);
      }
    }
  } else { // general theta
    for (unsigned int i = 0; i < n_strata; ++i) {
      mat Z_tv_theta_B1 =
        (Z_tv.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]) * theta) %
        repelem(B1.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]),
                          1, B2.n_cols);
      vec Z_ti_beta_ti;
      if (ti) {
        Z_ti_beta_ti =
          Z_ti.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]) * beta_ti;
      }
      mat B2_sp = B2.rows(idx_B_sp[i]);
      unsigned int n_Z_tv_theta = Z_tv_theta_B1.n_rows;
      vec lincomb_fail(idx_fail[i].size()), hazard_tmp(idx_fail[i].size());
      if (parallel) {
        omp_set_num_threads(threads);
        #pragma omp parallel
        {
          unsigned int id, j;
          id = omp_get_thread_num();
          double val_tmp = 0;
          if (ti) {
            for (j = istart[i][id]; j < iend[i][id]; ++j) {
              vec lincomb =
                sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
                repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
                Z_ti_beta_ti.subvec(idx_fail[i][j](0),n_Z_tv_theta-1);
              lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
              double tmp = sum(exp(lincomb));
              hazard_tmp(j) = idx_fail[i][j].n_elem/tmp;
              val_tmp += idx_fail[i][j].n_elem*log(tmp);
            }
          } else {
            for (j = istart[i][id]; j < iend[i][id]; ++j) {
              vec lincomb =
                sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
                repmat(B2_sp.row(j), 1, B1.n_cols), 1);
              lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
              double tmp = sum(exp(lincomb));
              hazard_tmp(j) = idx_fail[i][j].n_elem/tmp;
              val_tmp += idx_fail[i][j].n_elem*log(tmp);
            }
          }
          #pragma omp atomic
          logplkd -= val_tmp;
        }
      } else {
        if (ti) {
          for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
            unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1,
              nar = idx_fail[i][j].n_elem;
            vec lincomb =
              sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
              Z_ti_beta_ti.subvec(arstart,arend);
            lincomb_fail(j) = sum(lincomb.head(nar));
            double tmp = sum(exp(lincomb));
            hazard_tmp(j) = nar/tmp;
            logplkd -= nar*log(tmp);
          }
        } else {
          for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
            vec lincomb =
              sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1);
            lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
            double tmp = sum(exp(lincomb));
            hazard_tmp(j) = idx_fail[i][j].n_elem/tmp;
            logplkd -= idx_fail[i][j].n_elem*log(tmp);
          }
        }
      }
      logplkd += accu(lincomb_fail);
      hazard.push_back(hazard_tmp);
    }
  }
  logplkd /= Z_tv.n_rows;
  return List::create(_["logplkd"]=logplkd, _["hazard"]=hazard);
}

// [[Rcpp::export]]
double logplkd_fixtra_bresties_standalone(const vec &event, const vec &time, 
                                          const IntegerVector &count_strata,
                                          const mat &Z_tv, const mat &B1,
                                          const mat &B2, const mat &theta,
                                          const mat &Z_ti, const vec &beta_ti,
                                          const bool &parallel=false,
                                          const unsigned int &threads=1) {
  // theta: p by K_1K_2
  // B1: calendar time
  // B2: post-discharge time
  IntegerVector cumsum_strata = cumsum(count_strata);
  const unsigned int n_strata = cumsum_strata.length();
  cumsum_strata.push_front(0);
  vector<vector<uvec>> idx_fail;
  vector<vector<unsigned int>> idx_fail_1st, idx_Z_strata;
  vector<unsigned int> n_fail_time, n_Z_strata;
  // each element of idx_Z_strata contains start/end row indices of Z_strata
  for (unsigned int i = 0; i < n_strata; ++i) {
    vec event_tmp = event.rows(cumsum_strata[i], cumsum_strata[i+1]-1),
      time_tmp = time.rows(cumsum_strata[i], cumsum_strata[i+1]-1);
    uvec idx_fail_tmp = find(event_tmp==1);
    vector<unsigned int> idx_Z_strata_tmp;
    idx_Z_strata_tmp.push_back(cumsum_strata[i]+idx_fail_tmp(0));
    idx_Z_strata_tmp.push_back(cumsum_strata[i+1]-1);
    idx_Z_strata.push_back(idx_Z_strata_tmp);
    n_Z_strata.push_back(cumsum_strata[i+1]-cumsum_strata[i]-
      idx_fail_tmp(0));
    vec time_fail_tmp = time_tmp.elem(idx_fail_tmp);
    idx_fail_tmp -= idx_fail_tmp(0);
    vec uniq_t = unique(time_fail_tmp);
    n_fail_time.push_back(uniq_t.n_elem);
    vector<uvec> idx_fail_tmp_tmp;
    vector<unsigned int> idx_fail_1st_tmp;
    for (vec::iterator j = uniq_t.begin(); j < uniq_t.end(); ++j) {
      uvec tmp = idx_fail_tmp.elem(find(time_fail_tmp==*j));
      idx_fail_tmp_tmp.push_back(tmp);
      idx_fail_1st_tmp.push_back(tmp[0]);
    }
    idx_fail.push_back(idx_fail_tmp_tmp);
    idx_fail_1st.push_back(idx_fail_1st_tmp);
  }
  IntegerVector n_failtime = wrap(n_fail_time);
  IntegerVector cumsum_failtime = cumsum(n_failtime);
  cumsum_failtime.push_front(0);
  vector<uvec> idx_B_sp;
  for (unsigned int i = 0; i < n_strata; ++i) {
    idx_B_sp.push_back(regspace<uvec>(cumsum_failtime[i],
                                      cumsum_failtime[i+1]-1));
  }
  // istart and iend for each thread when parallel=true
  vector<vec> cumsum_ar;
  vector<vector<unsigned int>> istart, iend;
  if (parallel) {
    for (unsigned int i = 0; i < n_strata; ++i) {
      double scale_fac = as_scalar(idx_fail[i].back().tail(1));
      cumsum_ar.push_back(
        (double)n_Z_strata[i] / scale_fac * regspace(1, idx_fail[i].size()) -
          cumsum(conv_to<vec>::from(idx_fail_1st[i])/scale_fac));
      vector<unsigned int> istart_tmp, iend_tmp;
      for (unsigned int id = 0; id < threads; ++id) {
        istart_tmp.push_back(as_scalar(find(cumsum_ar[i] >=
          cumsum_ar[i](idx_fail[i].size()-1)/(double)threads*id, 1)));
        iend_tmp.push_back(as_scalar(find(cumsum_ar[i] >=
          cumsum_ar[i](idx_fail[i].size()-1)/(double)threads*(id+1), 1)));
        if (id == threads-1) {
          iend_tmp.pop_back();
          iend_tmp.push_back(idx_fail[i].size());
        }
      }
      istart.push_back(istart_tmp);
      iend.push_back(iend_tmp);
    }
  }
  
  bool ti = norm(Z_ti, "inf") > sqrt(datum::eps);
  double logplkd = 0.0;
  for (unsigned int i = 0; i < n_strata; ++i) {
    mat Z_tv_theta_B1 =
      (Z_tv.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]) * theta) %
      repelem(B1.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]),
              1, B2.n_cols);
    vec Z_ti_beta_ti;
    if (ti) {
      Z_ti_beta_ti =
        Z_ti.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]) * beta_ti;
    }
    mat B2_sp = B2.rows(idx_B_sp[i]);
    unsigned int n_Z_tv_theta = Z_tv_theta_B1.n_rows;
    vec lincomb_fail(idx_fail[i].size());
    if (parallel) {
      omp_set_num_threads(threads);
      #pragma omp parallel
      {
        unsigned int id, j;
        id = omp_get_thread_num();
        double val_tmp = 0;
        if (ti) {
          for (j = istart[i][id]; j < iend[i][id]; ++j) {
            vec lincomb =
              sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
              Z_ti_beta_ti.subvec(idx_fail[i][j](0),n_Z_tv_theta-1);
            lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
            val_tmp += idx_fail[i][j].n_elem*log(sum(exp(lincomb)));
          }
        } else {
          for (j = istart[i][id]; j < iend[i][id]; ++j) {
            vec lincomb =
              sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1);
            lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
            val_tmp += idx_fail[i][j].n_elem*log(sum(exp(lincomb)));
          }
        }
        #pragma omp atomic
          logplkd -= val_tmp;
      }
    } else {
      if (ti) {
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1,
            nar = idx_fail[i][j].n_elem;
          vec lincomb =
            sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
            repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
            Z_ti_beta_ti.subvec(arstart,arend);
          lincomb_fail(j) = sum(lincomb.head(nar));
          logplkd -= nar*log(sum(exp(lincomb)));
        }
      } else {
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          vec lincomb =
            sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
            repmat(B2_sp.row(j), 1, B1.n_cols), 1);
          lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
          logplkd -= idx_fail[i][j].n_elem*log(sum(exp(lincomb)));
        }
      }
    }
    logplkd += accu(lincomb_fail);
  }
  return logplkd;
}

// not used
mat info_fixtra_bresties_standalone(const mat &theta, const vec &event,
                                    const vec &time,
                                    const IntegerVector &count_strata,
                                    const mat &Z_tv, const mat &B1,
                                    const mat &B2, 
                                    const mat &Z_ti, const vec &beta_ti,
                                    const mat &D1, const mat &D2,
                                    const vec &mu1, const vec &mu2) {
  // data should be ordered by strata first and then time
  // theta: p by k; B2: t by k
  
  bool ti = norm(Z_ti, "inf") > sqrt(datum::eps);
  IntegerVector cumsum_strata = cumsum(count_strata);
  unsigned int n_strata = cumsum_strata.length();
  cumsum_strata.push_front(0);
  mat info;
  if (ti) {
    info = zeros<mat>(theta.n_elem+beta_ti.n_elem, theta.n_elem+beta_ti.n_elem);
  } else {
    info = zeros<mat>(theta.n_elem, theta.n_elem);
  }
  for (unsigned int i = 0; i < n_strata; ++i) {
    vec event_tmp = event.rows(cumsum_strata[i], cumsum_strata[i+1]-1),
      time_tmp = time.rows(cumsum_strata[i], cumsum_strata[i+1]-1);
    uvec idx_fail_tmp = find(event_tmp==1);
    vec time_fail_tmp = time_tmp.elem(idx_fail_tmp);
    unsigned int start = cumsum_strata[i]+idx_fail_tmp(0),
      end = cumsum_strata[i+1]-1;
    mat Z_tv_theta_B1 = (Z_tv.rows(start, end) * theta) %
      repelem(B1.rows(start, end), 1, B2.n_cols);
    mat Z_tv_B1 = repelem(Z_tv.rows(start, end), 1, B1.n_cols) %
      repmat(B1.rows(start, end), 1, Z_tv.n_cols);
    mat Z_ti_strata; vec Z_ti_beta_ti;
    if (ti) {
      Z_ti_strata = Z_ti.rows(start, end);
      Z_ti_beta_ti = Z_ti_strata * beta_ti;
    }
    mat B2_sp = B2.rows(cumsum_strata[i]+idx_fail_tmp);
    idx_fail_tmp -= idx_fail_tmp(0); // index within the strata
    vec uniq_t = unique(time_fail_tmp);
    vector<uvec> idx_fail;
    for (vec::iterator j = uniq_t.begin(); j < uniq_t.end(); ++j) {
      idx_fail.push_back(idx_fail_tmp.elem(find(time_fail_tmp==*j)));
    }
    unsigned int n_Z_tv_theta = Z_tv_theta_B1.n_rows;
    if (ti) {
      for (unsigned int j = 0; j < idx_fail.size(); ++j) {
        unsigned int arstart = idx_fail[j](0), arend = n_Z_tv_theta-1,
          nar = idx_fail[j].n_elem;
        vec exp_lincomb = exp(sum(Z_tv_theta_B1.rows(arstart,arend).each_row() %
          repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
          Z_ti_beta_ti.subvec(arstart,arend));
        double S0 = accu(exp_lincomb);
        mat Z_tv_tmp = kron(Z_tv_B1.rows(arstart,arend), B2_sp.row(j));
        mat Z_tmp = join_rows(Z_tv_tmp, Z_ti_strata.rows(arstart,arend));
        mat Z_exp_tmp = Z_tmp.each_col() % exp_lincomb;
        vec S1 = sum(Z_exp_tmp).t()/S0;
        mat S2 = Z_tmp.t() * Z_exp_tmp/S0;
        info += nar*(S2 - S1*S1.t());
      }
    } else {
      for (unsigned int j = 0; j < idx_fail.size(); ++j) {
        unsigned int arstart = idx_fail[j](0), arend = n_Z_tv_theta-1,
          nar = idx_fail[j].n_elem;
        vec exp_lincomb = exp(sum(Z_tv_theta_B1.rows(arstart,arend).each_row() %
          repmat(B2_sp.row(j), 1, B1.n_cols), 1));
                                double S0 = accu(exp_lincomb);
                                mat Z_tmp = kron(Z_tv_B1.rows(arstart,arend), B2_sp.row(j));
                                mat Z_exp_tmp = Z_tmp.each_col() % exp_lincomb;
                                vec S1 = sum(Z_exp_tmp).t()/S0;
                                mat S2 = Z_tmp.t() * Z_exp_tmp/S0;
                                info += nar*(S2 - S1*S1.t());
      }
    }
  }
  mat A = kron(kron(D1,eye(B2.n_cols,B2.n_cols)),diagmat(mu1)),
    B = kron(kron(eye(B1.n_cols,B1.n_cols),D2),diagmat(mu2));
  mat P = Z_tv.n_rows * (A.t() * A + B.t() * B);
  info.submat(0,0,theta.n_elem-1,theta.n_elem-1) += P;
  return info;
}

// [[Rcpp::export]]
double logplkd_fixtra_bresties_cv_UC(const vec &event, const vec &time, 
                                     const IntegerVector &count_strata,
                                     const uvec &fold,
                                     const mat &Z_tv, const mat &B1,
                                     const mat &B2, List &theta,
                                     const mat &Z_ti, List &beta_ti,
                                     const bool &parallel=false,
                                     const unsigned int &threads=1) {
  // theta: a list of thetas p by K_1K_2
  // B1: calendar time
  // B2: post-discharge time
  IntegerVector cumsum_strata = cumsum(count_strata);
  const unsigned int n_strata = cumsum_strata.length();
  cumsum_strata.push_front(0);
  vector<vector<uvec>> idx_fail;
  vector<vector<unsigned int>> idx_fail_1st, idx_Z_strata;
  vector<unsigned int> n_fail_time, n_Z_strata;
  // each element of idx_Z_strata contains start/end row indices of Z_strata
  for (unsigned int i = 0; i < n_strata; ++i) {
    vec event_tmp = event.rows(cumsum_strata[i], cumsum_strata[i+1]-1),
      time_tmp = time.rows(cumsum_strata[i], cumsum_strata[i+1]-1);
    uvec idx_fail_tmp = find(event_tmp==1);
    vector<unsigned int> idx_Z_strata_tmp;
    idx_Z_strata_tmp.push_back(cumsum_strata[i]+idx_fail_tmp(0));
    idx_Z_strata_tmp.push_back(cumsum_strata[i+1]-1);
    idx_Z_strata.push_back(idx_Z_strata_tmp);
    n_Z_strata.push_back(cumsum_strata[i+1]-cumsum_strata[i]-
      idx_fail_tmp(0));
    vec time_fail_tmp = time_tmp.elem(idx_fail_tmp);
    idx_fail_tmp -= idx_fail_tmp(0);
    vec uniq_t = unique(time_fail_tmp);
    n_fail_time.push_back(uniq_t.n_elem);
    vector<uvec> idx_fail_tmp_tmp;
    vector<unsigned int> idx_fail_1st_tmp;
    for (vec::iterator j = uniq_t.begin(); j < uniq_t.end(); ++j) {
      uvec tmp = idx_fail_tmp.elem(find(time_fail_tmp==*j));
      idx_fail_tmp_tmp.push_back(tmp);
      idx_fail_1st_tmp.push_back(tmp[0]);
    }
    idx_fail.push_back(idx_fail_tmp_tmp);
    idx_fail_1st.push_back(idx_fail_1st_tmp);
  }
  IntegerVector n_failtime = wrap(n_fail_time);
  IntegerVector cumsum_failtime = cumsum(n_failtime);
  cumsum_failtime.push_front(0);
  vector<uvec> idx_B_sp;
  for (unsigned int i = 0; i < n_strata; ++i) {
    idx_B_sp.push_back(regspace<uvec>(cumsum_failtime[i],
                                      cumsum_failtime[i+1]-1));
  }
  // istart and iend for each thread when parallel=true
  vector<vec> cumsum_ar;
  vector<vector<unsigned int>> istart, iend;
  if (parallel) {
    for (unsigned int i = 0; i < n_strata; ++i) {
      double scale_fac = as_scalar(idx_fail[i].back().tail(1));
      cumsum_ar.push_back(
        (double)n_Z_strata[i] / scale_fac * regspace(1, idx_fail[i].size()) -
          cumsum(conv_to<vec>::from(idx_fail_1st[i])/scale_fac));
      vector<unsigned int> istart_tmp, iend_tmp;
      for (unsigned int id = 0; id < threads; ++id) {
        istart_tmp.push_back(as_scalar(find(cumsum_ar[i] >=
          cumsum_ar[i](idx_fail[i].size()-1)/(double)threads*id, 1)));
        iend_tmp.push_back(as_scalar(find(cumsum_ar[i] >=
          cumsum_ar[i](idx_fail[i].size()-1)/(double)threads*(id+1), 1)));
        if (id == threads-1) {
          iend_tmp.pop_back();
          iend_tmp.push_back(idx_fail[i].size());
        }
      }
      istart.push_back(istart_tmp);
      iend.push_back(iend_tmp);
    }
  }
  
  bool ti = norm(Z_ti, "inf") > sqrt(datum::eps);
  double logplkd = 0.0;
  mat Z_tv_theta(Z_tv.n_rows, B1.n_cols*B2.n_cols);
  vec Z_ti_beta_ti_all(Z_ti.n_rows);
  for (unsigned int i = 0; i < max(fold); ++i) {
    mat theta_tmp = theta[i];
    Z_tv_theta.rows(find(fold==i+1)) = Z_tv.rows(find(fold==i+1)) * theta_tmp;
    if (ti) {
      vec beta_ti_tmp = beta_ti[i];
      Z_ti_beta_ti_all(find(fold==i+1)) = Z_ti.rows(find(fold==i+1)) * beta_ti_tmp;
    }
  }
  for (unsigned int i = 0; i < n_strata; ++i) {
    mat Z_tv_theta_B1 =
      Z_tv_theta.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]) %
      repelem(B1.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]),
              1, B2.n_cols);
    vec Z_ti_beta_ti;
    if (ti) {
      Z_ti_beta_ti = Z_ti_beta_ti_all.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]);
    }
    mat B2_sp = B2.rows(idx_B_sp[i]);
    unsigned int n_Z_tv_theta = Z_tv_theta_B1.n_rows;
    vec lincomb_fail(idx_fail[i].size());
    if (parallel) {
      omp_set_num_threads(threads);
      #pragma omp parallel
      {
        unsigned int id, j;
        id = omp_get_thread_num();
        double val_tmp = 0;
        if (ti) {
          for (j = istart[i][id]; j < iend[i][id]; ++j) {
            vec lincomb =
              sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
              Z_ti_beta_ti.subvec(idx_fail[i][j](0),n_Z_tv_theta-1);
            lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
            val_tmp += idx_fail[i][j].n_elem*log(sum(exp(lincomb)));
          }
        } else {
          for (j = istart[i][id]; j < iend[i][id]; ++j) {
            vec lincomb =
              sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1);
            lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
            val_tmp += idx_fail[i][j].n_elem*log(sum(exp(lincomb)));
          }
        }
      #pragma omp atomic
        logplkd -= val_tmp;
}
    } else {
      if (ti) {
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1,
            nar = idx_fail[i][j].n_elem;
          vec lincomb =
            sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
            repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
            Z_ti_beta_ti.subvec(arstart,arend);
          lincomb_fail(j) = sum(lincomb.head(nar));
          logplkd -= nar*log(sum(exp(lincomb)));
        }
      } else {
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          vec lincomb =
            sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
            repmat(B2_sp.row(j), 1, B1.n_cols), 1);
          lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
          logplkd -= idx_fail[i][j].n_elem*log(sum(exp(lincomb)));
        }
      }
    }
    logplkd += accu(lincomb_fail);
  }
  return logplkd;
}

// [[Rcpp::export]]
double sumDRsq_fixtra_bresties_cv_DR(const vec &event, const vec &time, 
                                     const IntegerVector &count_strata,
                                     const uvec &fold,
                                     const mat &Z_tv, const mat &B1,
                                     const mat &B2, List &theta,
                                     const mat &Z_ti, List &beta_ti,
                                     const vector<vec> hazard) {
  // theta: a list of thetas p by K_1K_2
  // B1: calendar time
  // B2: post-discharge time
  IntegerVector cumsum_strata = cumsum(count_strata);
  const unsigned int n_strata = cumsum_strata.length();
  cumsum_strata.push_front(0);
  vector<vector<uvec>> idx_fail;
  vector<vector<unsigned int>> idx_fail_1st, idx_Z_strata;
  vector<unsigned int> n_fail_time, n_Z_strata;
  // each element of idx_Z_strata contains start/end row indices of Z_strata
  for (unsigned int i = 0; i < n_strata; ++i) {
    vec event_tmp = event.rows(cumsum_strata[i], cumsum_strata[i+1]-1),
      time_tmp = time.rows(cumsum_strata[i], cumsum_strata[i+1]-1);
    uvec idx_fail_tmp = find(event_tmp==1);
    vector<unsigned int> idx_Z_strata_tmp;
    idx_Z_strata_tmp.push_back(cumsum_strata[i]+idx_fail_tmp(0));
    idx_Z_strata_tmp.push_back(cumsum_strata[i+1]-1);
    idx_Z_strata.push_back(idx_Z_strata_tmp);
    n_Z_strata.push_back(cumsum_strata[i+1]-cumsum_strata[i]-
      idx_fail_tmp(0));
    vec time_fail_tmp = time_tmp.elem(idx_fail_tmp);
    idx_fail_tmp -= idx_fail_tmp(0);
    vec uniq_t = unique(time_fail_tmp);
    n_fail_time.push_back(uniq_t.n_elem);
    vector<uvec> idx_fail_tmp_tmp;
    vector<unsigned int> idx_fail_1st_tmp;
    for (vec::iterator j = uniq_t.begin(); j < uniq_t.end(); ++j) {
      uvec tmp = idx_fail_tmp.elem(find(time_fail_tmp==*j));
      idx_fail_tmp_tmp.push_back(tmp);
      idx_fail_1st_tmp.push_back(tmp[0]);
    }
    idx_fail.push_back(idx_fail_tmp_tmp);
    idx_fail_1st.push_back(idx_fail_1st_tmp);
  }
  IntegerVector n_failtime = wrap(n_fail_time);
  IntegerVector cumsum_failtime = cumsum(n_failtime);
  cumsum_failtime.push_front(0);
  vector<uvec> idx_B_sp;
  for (unsigned int i = 0; i < n_strata; ++i) {
    idx_B_sp.push_back(regspace<uvec>(cumsum_failtime[i],
                                      cumsum_failtime[i+1]-1));
  }
  bool ti = norm(Z_ti, "inf") > sqrt(datum::eps);
  double obj = 0.0;
  mat Z_tv_theta(Z_tv.n_rows, B1.n_cols*B2.n_cols);
  vec Z_ti_beta_ti_all(Z_ti.n_rows);
  for (unsigned int i = 0; i < max(fold); ++i) {
    mat theta_tmp = theta[i];
    Z_tv_theta.rows(find(fold==i+1)) = Z_tv.rows(find(fold==i+1)) * theta_tmp;
    if (ti) {
      vec beta_ti_tmp = beta_ti[i];
      Z_ti_beta_ti_all(find(fold==i+1)) = Z_ti.rows(find(fold==i+1)) * beta_ti_tmp;
    }
  }
  for (unsigned int i = 0; i < n_strata; ++i) {
    mat Z_tv_theta_B1 =
      Z_tv_theta.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]) %
      repelem(B1.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]),
              1, B2.n_cols);
    vec Z_ti_beta_ti;
    if (ti) {
      Z_ti_beta_ti = Z_ti_beta_ti_all.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]);
    }
    mat B2_sp = B2.rows(idx_B_sp[i]);
    unsigned int n_Z_tv_theta = Z_tv_theta_B1.n_rows;
    vec lincomb_fail(idx_fail[i].size());
      double cumhazard = 0.0;
      if (ti) {
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1,
            nar = idx_fail[i][j].n_elem;
          cumhazard += hazard[i](j);
          vec lincomb =
            sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
            repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
            Z_ti_beta_ti.subvec(arstart,arend);
          lincomb_fail(j) = sum(lincomb.head(nar)) + nar + nar*log(cumhazard);
          obj -= nar*sum(exp(lincomb)*hazard[i](j));
        }
      } else {
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          unsigned int nar = idx_fail[i][j].n_elem;
          cumhazard += hazard[i](j);
          vec lincomb =
            sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
            repmat(B2_sp.row(j), 1, B1.n_cols), 1);
          lincomb_fail(j) = sum(lincomb.head(nar)) + nar + nar*log(cumhazard);
          obj -= nar*sum(exp(lincomb)*hazard[i](j));
        }
      }
    obj += accu(lincomb_fail);
  }
  return -2*obj;
}

List logplkd_fixtra_bresties_diffpen(const mat &Z_tv,
                                     const mat &B1, const mat &B2,
                                     const mat &theta,
                                     const mat &Z_ti, const vec &beta_ti,
                                     const bool &ti,
                                     const unsigned int n_strata,
                                     const mat &D1, const mat &D2,
                                     const vec &mu1, const vec &mu2,
                                     vector<uvec> &idx_B_sp,
                                     vector<vector<uvec>> &idx_fail,
                                     vector<unsigned int> n_Z_strata,
                                     vector<vector<unsigned int>> &idx_Z_strata,
                                     vector<vector<unsigned int>> &istart,
                                     vector<vector<unsigned int>> &iend,
                                     const bool &parallel=false,
                                     const unsigned int &threads=1) {
  // theta: p by K_1K_2
  // B1: calendar time
  // B2: post-discharge time
  double norm_parm;
  vector<vec> hazard;
  if (ti) {
    norm_parm = max(norm(theta, "inf"), norm(beta_ti, "inf"));
  } else {
    norm_parm = norm(theta, "inf");
  }
  double logplkd = 0.0;
  if (norm_parm < sqrt(datum::eps)) { // theta and beta_ti are 0
    if (parallel) {
      for (unsigned int i = 0; i < n_strata; ++i) {
        vec hazard_tmp(idx_fail[i].size());
        omp_set_num_threads(threads);
        #pragma omp parallel
        {
          unsigned int id, j;
          id = omp_get_thread_num();
          double val_tmp = 0;
          for (j = istart[i][id]; j < iend[i][id]; ++j) {
            double tmp = n_Z_strata[i]-idx_fail[i][j](0);
            hazard_tmp(j) = idx_fail[i][j].n_elem/tmp;
            val_tmp += idx_fail[i][j].n_elem*log(tmp);
          }
        #pragma omp atomic
          logplkd -= val_tmp;
        }
        hazard.push_back(hazard_tmp);
      }
    } else {
      for (unsigned int i = 0; i < n_strata; ++i) {
        vec hazard_tmp(idx_fail[i].size());
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          double tmp = n_Z_strata[i]-idx_fail[i][j](0);
          hazard_tmp(j) = idx_fail[i][j].n_elem/tmp;
          logplkd -= idx_fail[i][j].n_elem*log(tmp);
        }
        hazard.push_back(hazard_tmp);
      }
    }
  } else { // general theta
    for (unsigned int i = 0; i < n_strata; ++i) {
      mat Z_tv_theta_B1 =
        (Z_tv.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]) * theta) %
        repelem(B1.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]),
                1, B2.n_cols);
      vec Z_ti_beta_ti;
      if (ti) {
        Z_ti_beta_ti =
          Z_ti.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]) * beta_ti;
      }
      mat B2_sp = B2.rows(idx_B_sp[i]);
      unsigned int n_Z_tv_theta = Z_tv_theta_B1.n_rows;
      vec lincomb_fail(idx_fail[i].size()), hazard_tmp(idx_fail[i].size());
      if (parallel) {
        omp_set_num_threads(threads);
        #pragma omp parallel
        {
          unsigned int id, j;
          id = omp_get_thread_num();
          double val_tmp = 0;
          if (ti) {
            for (j = istart[i][id]; j < iend[i][id]; ++j) {
              vec lincomb =
                sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
                repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
                Z_ti_beta_ti.subvec(idx_fail[i][j](0),n_Z_tv_theta-1);
              lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
              double tmp = sum(exp(lincomb));
              hazard_tmp(j) = idx_fail[i][j].n_elem/tmp;
              val_tmp += idx_fail[i][j].n_elem*log(tmp);
            }
          } else {
            for (j = istart[i][id]; j < iend[i][id]; ++j) {
              vec lincomb =
                sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
                repmat(B2_sp.row(j), 1, B1.n_cols), 1);
              lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
              double tmp = sum(exp(lincomb));
              hazard_tmp(j) = idx_fail[i][j].n_elem/tmp;
              val_tmp += idx_fail[i][j].n_elem*log(tmp);
            }
          }
        #pragma omp atomic
          logplkd -= val_tmp;
        }
      } else {
        if (ti) {
          for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
            unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1,
              nar = idx_fail[i][j].n_elem;
            vec lincomb =
              sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
              Z_ti_beta_ti.subvec(arstart,arend);
            lincomb_fail(j) = sum(lincomb.head(nar));
            double tmp = sum(exp(lincomb));
            hazard_tmp(j) = nar/tmp;
            logplkd -= nar*log(tmp);
          }
        } else {
          for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
            vec lincomb =
              sum(Z_tv_theta_B1.rows(idx_fail[i][j](0), n_Z_tv_theta-1).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1);
            lincomb_fail(j) = sum(lincomb.head(idx_fail[i][j].n_elem));
            double tmp = sum(exp(lincomb));
            hazard_tmp(j) = idx_fail[i][j].n_elem/tmp;
            logplkd -= idx_fail[i][j].n_elem*log(tmp);
          }
        }
      }
      logplkd += accu(lincomb_fail);
      hazard.push_back(hazard_tmp);
    }
  }
  logplkd -= (pow(norm(kron(D1,eye(B2.n_cols,B2.n_cols))*theta.t()*diagmat(mu1), "fro"),2) +
    pow(norm(kron(eye(B1.n_cols,B1.n_cols), D2)*theta.t()*diagmat(mu2), "fro"),2)) * pow(Z_tv.n_rows, 1);
  logplkd /= Z_tv.n_rows;
  return List::create(_["logplkd"]=logplkd, _["hazard"]=hazard);
}

List stepinc_fixtra_bresties(const mat &Z_tv, const mat &B1,
                             const mat &B2,
                             const mat &theta, const mat &Z_ti,
                             const vec &beta_ti, const bool &ti,
                             const unsigned int n_strata,
                             vector<uvec> &idx_B_sp,
                             vector<vector<uvec>> &idx_fail,
                             vector<vector<unsigned int>> &idx_Z_strata,
                             vector<vector<unsigned int>> &istart,
                             vector<vector<unsigned int>> &iend,
                             const string &method="Newton",
                             const double &lambda=1e8,
                             const bool &parallel=false,
                             const unsigned int &threads=1) {
  vec grad; mat info; // gradient and info matrix
  if (ti) {
    grad = zeros<vec>(theta.n_elem+beta_ti.n_elem);
    info = zeros<mat>(theta.n_elem+beta_ti.n_elem, theta.n_elem+beta_ti.n_elem);
  } else {
    grad = zeros<vec>(theta.n_elem);
    info = zeros<mat>(theta.n_elem, theta.n_elem);
  }
  for (unsigned int i = 0; i < n_strata; ++i) {
    mat Z_tv_theta_B1 =
      (Z_tv.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]) * theta) %
      repelem(B1.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]),
              1, B2.n_cols);
    mat Z_tv_B1 =
      repelem(Z_tv.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]), 1, B1.n_cols) %
      repmat(B1.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]), 1, Z_tv.n_cols);
    // mat Z_tv_tmp = Z_tv.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]);
    // mat B1_sp = B1.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]);
    // mat Z_tv_B1_2(B1_sp.n_rows, Z_tv_tmp.n_cols*B1_sp.n_cols, fill::zeros);
    // for (unsigned int k = 0; k < B1_sp.n_rows; ++k) {
    //   Z_tv_B1_2.row(k) += kron(Z_tv_tmp.row(k), B1_sp.row(k));
    // }
    // Rcout << "colmean of Z_tv_B1 =" << mean(Z_tv_B1) << endl;
    // Rcout << "colmean of Z_tv_B1_2 =" << mean(Z_tv_B1_2) << endl;
    mat Z_ti_strata; vec Z_ti_beta_ti;
    if (ti) {
      Z_ti_strata = Z_ti.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]);
      Z_ti_beta_ti = Z_ti_strata * beta_ti;
    }
    mat B2_sp = B2.rows(idx_B_sp[i]);
    // if (i == 10) {
    //   Rcout << accu(Z_tv_theta_B1) << endl;
    //   
    //   Rcout << accu(Z_tv_B1) << endl;
    // }
    unsigned int n_Z_tv_theta = Z_tv_theta_B1.n_rows;
    if (parallel) {
      omp_set_num_threads(threads);
      #pragma omp parallel
      {
        unsigned int id, j;
        id = omp_get_thread_num();
        vec grad_tmp(size(grad), fill::zeros);
        mat info_tmp(size(info), fill::zeros);
        if (ti) {
          for (j = istart[i][id]; j < iend[i][id]; ++j) {
            unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1, 
            nar = idx_fail[i][j].n_elem;
            vec exp_lincomb = exp(
              sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
                repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
                Z_ti_beta_ti.subvec(arstart,arend));
            double S0 = accu(exp_lincomb);
            mat Z_tmp = join_rows(kron(Z_tv_B1.rows(arstart,arend), B2_sp.row(j)),
                                  Z_ti_strata.rows(arstart,arend));
            mat Z_exp_tmp = Z_tmp.each_col() % exp_lincomb;
            vec S1 = sum(Z_exp_tmp).t()/S0;
            mat S2 = Z_tmp.t() * Z_exp_tmp/S0;
            grad_tmp += sum(Z_tmp.rows(idx_fail[i][j]-arstart)).t() - nar*S1;
            info_tmp += nar*(S2 - S1*S1.t());
          }
        } else {
          for (j = istart[i][id]; j < iend[i][id]; ++j) {
            unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1,
            nar = idx_fail[i][j].n_elem;
            vec exp_lincomb = exp(
              sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
                repmat(B2_sp.row(j), 1, B1.n_cols), 1));
            double S0 = accu(exp_lincomb);
            mat Z_tmp = kron(Z_tv_B1.rows(arstart,arend), B2_sp.row(j));
            mat Z_exp_tmp = Z_tmp.each_col() % exp_lincomb;
            vec S1 = sum(Z_exp_tmp).t()/S0;
            mat S2 = Z_tmp.t() * Z_exp_tmp/S0;
            grad_tmp += sum(Z_tmp.rows(idx_fail[i][j]-arstart)).t() - nar*S1;
            info_tmp += nar*(S2 - S1*S1.t());
          }
        }
        #pragma omp critical (gradient)
        grad += grad_tmp;
        #pragma omp critical (information)
        info += info_tmp;
      }
    } else {
      if (ti) {
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1,
            nar = idx_fail[i][j].n_elem;
          vec exp_lincomb = exp(
            sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
              Z_ti_beta_ti.subvec(arstart,arend));
          double S0 = accu(exp_lincomb);
          mat Z_tv_tmp = kron(Z_tv_B1.rows(arstart,arend), B2_sp.row(j));
          mat Z_tmp = join_rows(Z_tv_tmp, Z_ti_strata.rows(arstart,arend));
          mat Z_exp_tmp = Z_tmp.each_col() % exp_lincomb;
          vec S1 = sum(Z_exp_tmp).t()/S0;
          mat S2 = Z_tmp.t() * Z_exp_tmp/S0;
          grad += sum(Z_tmp.rows(idx_fail[i][j]-arstart)).t() - nar*S1;
          info += nar*(S2 - S1*S1.t());
          // if (j==18) {
          // //   // Rcout << "Z_tv_B1 = " << Z_tv_B1.row(arend) << endl;
          // Rcout << "inc grad = " << Z_tmp.rows(idx_fail[i][j]-arstart) << endl;
          // // Rcout << "sum S1 = " << accu(S1) << ", sum S2 = " << accu(S2) << endl;
          // //   Rcout << "arstart = " << arstart << ", arend =  " << arend << endl;
          // //   Rcout << "Z_tmp ncol = " << Z_tmp.n_cols << endl;
          // //   Rcout << "idx_fail[i][j] = " << idx_fail[i][j] << endl;
          // //   Rcout << "Z_tv_tmp = " << Z_tv_tmp.rows(idx_fail[i][j]) << endl;
          // //   Rcout << "nrow of Z_tv_tmp = " << Z_tv_tmp.n_rows << endl;
          // //   Rcout << "part 12 = " << Z_ti_strata.row(arstart) << endl;
          // //   Rcout << "kron = " << kron(Z_tv_B1.row(arstart), B2_sp.row(j)) << endl;
          // //   // Rcout << "kron0 = " << kron(Z_tv_B1.row(arend-1), B2_sp.row(j)) << endl;
          // //   // Rcout << "kron1 = " << kron(Z_tv_B1.row(arend), B2_sp.row(j)) << endl;
          // }
        }

      } else {
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1,
          nar = idx_fail[i][j].n_elem;
          vec exp_lincomb = exp(
            sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1));
          double S0 = accu(exp_lincomb);
          mat Z_tmp = kron(Z_tv_B1.rows(arstart,arend), B2_sp.row(j));
          mat Z_exp_tmp = Z_tmp.each_col() % exp_lincomb;
          vec S1 = sum(Z_exp_tmp).t()/S0;
          mat S2 = Z_tmp.t() * Z_exp_tmp/S0;
          grad += sum(Z_tmp.rows(idx_fail[i][j]-arstart)).t() - nar*S1;
          info += nar*(S2 - S1*S1.t());
        }
      }
    }
    // Rcout << "sum grad = " << accu(grad) << endl;
  }
  vec step; // Newton step
  if (method=="ProxN") {
    info.diag() += Z_tv.n_rows / lambda;
    step = solve(info, grad, solve_opts::fast+solve_opts::likely_sympd);
    info.diag() -= Z_tv.n_rows / lambda;
  } else if (method=="Newton") {
    step = solve(info, grad, solve_opts::fast+solve_opts::likely_sympd);
  }
  double inc = dot(grad, step) / Z_tv.n_rows; // square of Newton increment
  if (ti) {
    return List::create(_["step_tv"]=reshape(step.head(theta.n_elem),
                                             size(theta.t())).t(),
                         _["step_ti"]=step.tail(beta_ti.n_elem),
                         _["grad"]=grad, _["info"]=info, _["inc"]=inc);
  } else {
    return List::create(_["step"]=reshape(step, size(theta.t())).t(),
                         _["grad"]=grad, _["info"]=info, _["inc"]=inc);
  }
}

List stepinc_fixtra_bresties_diffpen(const mat &Z_tv, const mat &B1,
                                     const mat &B2,
                                     const mat &theta, const mat &Z_ti,
                                     const vec &beta_ti, const bool &ti,
                                     const unsigned int n_strata,
                                     const mat &D1, const mat &D2,
                                     const vec &mu1, const vec &mu2,
                                     vector<uvec> &idx_B_sp,
                                     vector<vector<uvec>> &idx_fail,
                                     vector<vector<unsigned int>> &idx_Z_strata,
                                     vector<vector<unsigned int>> &istart,
                                     vector<vector<unsigned int>> &iend,
                                     const string &method="Newton",
                                     const double &lambda=1e8,
                                     const bool &parallel=false,
                                     const unsigned int &threads=1) {
  vec grad; mat info; // gradient and info matrix
  if (ti) {
    grad = zeros<vec>(theta.n_elem+beta_ti.n_elem);
    info = zeros<mat>(theta.n_elem+beta_ti.n_elem, theta.n_elem+beta_ti.n_elem);
  } else {
    grad = zeros<vec>(theta.n_elem);
    info = zeros<mat>(theta.n_elem, theta.n_elem);
  }
  for (unsigned int i = 0; i < n_strata; ++i) {
    mat Z_tv_theta_B1 =
      (Z_tv.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]) * theta) %
      repelem(B1.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]),
              1, B2.n_cols);
    mat Z_tv_B1 =
      repelem(Z_tv.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]), 1, B1.n_cols) %
      repmat(B1.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]), 1, Z_tv.n_cols);
    mat Z_ti_strata; vec Z_ti_beta_ti;
    if (ti) {
      Z_ti_strata = Z_ti.rows(idx_Z_strata[i][0], idx_Z_strata[i][1]);
      Z_ti_beta_ti = Z_ti_strata * beta_ti;
    }
    mat B2_sp = B2.rows(idx_B_sp[i]);
    unsigned int n_Z_tv_theta = Z_tv_theta_B1.n_rows;
    if (parallel) {
      omp_set_num_threads(threads);
      #pragma omp parallel
      {
        unsigned int id, j;
        id = omp_get_thread_num();
        vec grad_tmp(size(grad), fill::zeros);
        mat info_tmp(size(info), fill::zeros);
        if (ti) {
          for (j = istart[i][id]; j < iend[i][id]; ++j) {
            unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1, 
              nar = idx_fail[i][j].n_elem;
            vec exp_lincomb = exp(
              sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
                repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
                Z_ti_beta_ti.subvec(arstart,arend));
            double S0 = accu(exp_lincomb);
            mat Z_tmp = join_rows(kron(Z_tv_B1.rows(arstart,arend), B2_sp.row(j)),
                                  Z_ti_strata.rows(arstart,arend));
            mat Z_exp_tmp = Z_tmp.each_col() % exp_lincomb;
            vec S1 = sum(Z_exp_tmp).t()/S0;
            mat S2 = Z_tmp.t() * Z_exp_tmp/S0;
            grad_tmp += sum(Z_tmp.rows(idx_fail[i][j]-arstart)).t() - nar*S1;
            info_tmp += nar*(S2 - S1*S1.t());
          }
        } else {
          for (j = istart[i][id]; j < iend[i][id]; ++j) {
            unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1,
              nar = idx_fail[i][j].n_elem;
            vec exp_lincomb = exp(
              sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
                repmat(B2_sp.row(j), 1, B1.n_cols), 1));
            double S0 = accu(exp_lincomb);
            mat Z_tmp = kron(Z_tv_B1.rows(arstart,arend), B2_sp.row(j));
            mat Z_exp_tmp = Z_tmp.each_col() % exp_lincomb;
            vec S1 = sum(Z_exp_tmp).t()/S0;
            mat S2 = Z_tmp.t() * Z_exp_tmp/S0;
            grad_tmp += sum(Z_tmp.rows(idx_fail[i][j]-arstart)).t() - nar*S1;
            info_tmp += nar*(S2 - S1*S1.t());
          }
        }
      #pragma omp critical (gradient)
        grad += grad_tmp;
      #pragma omp critical (information)
        info += info_tmp;
      }
    } else {
      if (ti) {
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1,
            nar = idx_fail[i][j].n_elem;
          vec exp_lincomb = exp(
            sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1) +
              Z_ti_beta_ti.subvec(arstart,arend));
          double S0 = accu(exp_lincomb);
          mat Z_tv_tmp = kron(Z_tv_B1.rows(arstart,arend), B2_sp.row(j));
          mat Z_tmp = join_rows(Z_tv_tmp, Z_ti_strata.rows(arstart,arend));
          mat Z_exp_tmp = Z_tmp.each_col() % exp_lincomb;
          vec S1 = sum(Z_exp_tmp).t()/S0;
          mat S2 = Z_tmp.t() * Z_exp_tmp/S0;
          grad += sum(Z_tmp.rows(idx_fail[i][j]-arstart)).t() - nar*S1;
          info += nar*(S2 - S1*S1.t());
        }
      } else {
        for (unsigned int j = 0; j < idx_fail[i].size(); ++j) {
          unsigned int arstart = idx_fail[i][j](0), arend = n_Z_tv_theta-1,
            nar = idx_fail[i][j].n_elem;
          vec exp_lincomb = exp(
            sum(Z_tv_theta_B1.rows(arstart, arend).each_row() %
              repmat(B2_sp.row(j), 1, B1.n_cols), 1));
          double S0 = accu(exp_lincomb);
          mat Z_tmp = kron(Z_tv_B1.rows(arstart,arend), B2_sp.row(j));
          mat Z_exp_tmp = Z_tmp.each_col() % exp_lincomb;
          vec S1 = sum(Z_exp_tmp).t()/S0;
          mat S2 = Z_tmp.t() * Z_exp_tmp/S0;
          grad += sum(Z_tmp.rows(idx_fail[i][j]-arstart)).t() - nar*S1;
          info += nar*(S2 - S1*S1.t());
        }
      }
    }
  }
  vec step; // Newton step
  mat A = kron(kron(D1,eye(B2.n_cols,B2.n_cols)),diagmat(mu1)),
    B = kron(kron(eye(B1.n_cols,B1.n_cols),D2),diagmat(mu2));
  mat P = pow(Z_tv.n_rows, 1) * (A.t() * A + B.t() * B);
  grad.head(theta.n_elem) -= P * vectorise(theta.t());
  info.submat(0,0,theta.n_elem-1,theta.n_elem-1) += P;
  if (method=="ProxN") {
    info.diag() += Z_tv.n_rows / lambda;
    step = solve(info, grad, solve_opts::fast+solve_opts::likely_sympd);
    info.diag() -= Z_tv.n_rows / lambda;
  } else if (method=="Newton") {
    step = solve(info, grad, solve_opts::fast+solve_opts::likely_sympd);
  }
  double inc = dot(grad, step) / Z_tv.n_rows; // square of Newton increment
  if (ti) {
    return List::create(_["step_tv"]=reshape(step.head(theta.n_elem),
                                      size(theta.t())).t(),
                                      _["step_ti"]=step.tail(beta_ti.n_elem),
                                      _["grad"]=grad, _["info"]=info, _["inc"]=inc);
  } else {
    return List::create(_["step"]=reshape(step, size(theta.t())).t(),
                        _["grad"]=grad, _["info"]=info, _["inc"]=inc);
  }
}

// [[Rcpp::export]]
List surtiver_fixtra_bresties_fit(const vec &event, const vec &time, 
                                  const IntegerVector &count_strata,
                                  const mat &Z_tv, const mat &B1,
                                  const mat &B2,
                                  const mat &theta_init,
                                  const mat &Z_ti, const vec &beta_ti_init,
                                  const string &method="Newton",
                                  const double lambda=1e8, 
                                  const double &factor=1.0,
                                  const bool &parallel=false, 
                                  const unsigned int &threads=1,
                                  const double &tol=1e-10,
                                  const unsigned int &iter_max=20,
                                  const double &s=1e-2, const double &t=0.6,
                                  const string &btr="dynamic",
                                  const string &stop="incre") {
  // B1 has the same number of rows as Z_tv and Z_ti
  // B2 only includes distinct failure times sorted by strata
  bool ti = norm(Z_ti, "inf") > sqrt(datum::eps);
  IntegerVector cumsum_strata = cumsum(count_strata);
  const unsigned int n_strata = cumsum_strata.length();
  cumsum_strata.push_front(0);
  vector<vector<uvec>> idx_fail;
  vector<vector<unsigned int>> idx_fail_1st, idx_Z_strata;
  vector<unsigned int> n_fail_time, n_Z_strata;
  // each element of idx_Z_strata contains start/end row indices of Z_strata
  for (unsigned int i = 0; i < n_strata; ++i) {
    vec event_tmp = event.rows(cumsum_strata[i], cumsum_strata[i+1]-1),
    time_tmp = time.rows(cumsum_strata[i], cumsum_strata[i+1]-1);
    uvec idx_fail_tmp = find(event_tmp==1);
    vector<unsigned int> idx_Z_strata_tmp;
    idx_Z_strata_tmp.push_back(cumsum_strata[i]+idx_fail_tmp(0));
    idx_Z_strata_tmp.push_back(cumsum_strata[i+1]-1);
    idx_Z_strata.push_back(idx_Z_strata_tmp);
    n_Z_strata.push_back(cumsum_strata[i+1]-cumsum_strata[i]-
                           idx_fail_tmp(0));
    vec time_fail_tmp = time_tmp.elem(idx_fail_tmp);
    idx_fail_tmp -= idx_fail_tmp(0);
    vec uniq_t = unique(time_fail_tmp);
    n_fail_time.push_back(uniq_t.n_elem);
    vector<uvec> idx_fail_tmp_tmp;
    vector<unsigned int> idx_fail_1st_tmp;
    for (vec::iterator j = uniq_t.begin(); j < uniq_t.end(); ++j) {
      uvec tmp = idx_fail_tmp.elem(find(time_fail_tmp==*j));
      idx_fail_tmp_tmp.push_back(tmp);
      idx_fail_1st_tmp.push_back(tmp[0]);
    }
    idx_fail.push_back(idx_fail_tmp_tmp);
    idx_fail_1st.push_back(idx_fail_1st_tmp);
  }
  IntegerVector n_failtime = wrap(n_fail_time);
  IntegerVector cumsum_failtime = cumsum(n_failtime);
  cumsum_failtime.push_front(0);
  vector<uvec> idx_B_sp;
  for (unsigned int i = 0; i < n_strata; ++i) {
    idx_B_sp.push_back(regspace<uvec>(cumsum_failtime[i],
                                      cumsum_failtime[i+1]-1));
  }
  // istart and iend for each thread when parallel=true
  vector<vec> cumsum_ar;
  vector<vector<unsigned int>> istart, iend;
  if (parallel) {
    for (unsigned int i = 0; i < n_strata; ++i) {
      double scale_fac = as_scalar(idx_fail[i].back().tail(1));
      cumsum_ar.push_back(
        (double)n_Z_strata[i] / scale_fac * regspace(1, idx_fail[i].size()) -
          cumsum(conv_to<vec>::from(idx_fail_1st[i])/scale_fac));
      vector<unsigned int> istart_tmp, iend_tmp;
      for (unsigned int id = 0; id < threads; ++id) {
        istart_tmp.push_back(as_scalar(find(cumsum_ar[i] >=
                                              cumsum_ar[i](idx_fail[i].size()-1)/(double)threads*id, 1)));
        iend_tmp.push_back(as_scalar(find(cumsum_ar[i] >=
                                            cumsum_ar[i](idx_fail[i].size()-1)/(double)threads*(id+1), 1)));
        if (id == threads-1) {
          iend_tmp.pop_back();
          iend_tmp.push_back(idx_fail[i].size());
        }
      }
      istart.push_back(istart_tmp);
      iend.push_back(iend_tmp);
    }
  }
  // iterative algorithm with backtracking line search
  unsigned int iter = 0, btr_max = 100, btr_ct = 0;
  mat theta = theta_init; vec beta_ti = beta_ti_init;
  List theta_list = List::create(theta), beta_ti_list = List::create(beta_ti);
  double crit = 1.0, v = 1.0, objfun_init = 0, objfun, diff_objfun,
  inc, rhs_btr = 0;
  List logplkd_list, update_list;
  NumericVector objfun_vec;
  logplkd_list = logplkd_fixtra_bresties(Z_tv, B1, B2, theta, Z_ti, beta_ti,
                                         ti, n_strata, idx_B_sp, idx_fail,
                                         n_Z_strata, idx_Z_strata,
                                         istart, iend, parallel, threads);
  objfun = logplkd_list["logplkd"];
  objfun_vec.push_back(objfun);
  while (iter < iter_max && crit > tol) {
    ++iter;
    // if (iter > 1) break;
    update_list = stepinc_fixtra_bresties(Z_tv, B1, B2, theta, Z_ti, beta_ti,
                                          ti, n_strata, idx_B_sp, idx_fail,
                                          idx_Z_strata, istart, iend, method,
                                          lambda*pow(factor,iter-1),
                                          parallel, threads);
    v = 1.0; // reset step size
    if (ti) {
      mat step_tv = update_list["step_tv"];
      vec step_ti = update_list["step_ti"];
      inc = update_list["inc"];
      // if (iter==1) {
      //   Rcout << inc << endl;
      // }
      if (btr=="none") {
        theta += step_tv;
        beta_ti += step_ti;
        theta_list.push_back(theta); beta_ti_list.push_back(beta_ti);
        crit = inc / 2;
        logplkd_list = logplkd_fixtra_bresties(Z_tv, B1, B2, theta, Z_ti, beta_ti,
                                               ti, n_strata, idx_B_sp, idx_fail,
                                               n_Z_strata, idx_Z_strata,
                                               istart, iend, parallel, threads);
        objfun = logplkd_list["logplkd"];
      } else {
        mat theta_tmp = theta + step_tv;
        vec beta_ti_tmp = beta_ti + step_ti;
        logplkd_list = logplkd_fixtra_bresties(Z_tv, B1, B2, theta_tmp, Z_ti,
                                               beta_ti_tmp, ti, n_strata, idx_B_sp,
                                               idx_fail, n_Z_strata, idx_Z_strata,
                                               istart, iend, parallel, threads);
        double objfun_tmp = logplkd_list["logplkd"];
        diff_objfun = objfun_tmp - objfun;
        if (btr=="dynamic")      rhs_btr = inc;
        else if (btr=="static")  rhs_btr = 1.0;
        while (diff_objfun < s * v * rhs_btr && btr_ct < btr_max) {
          ++btr_ct;
          v *= t;
          theta_tmp = theta + v * step_tv;
          beta_ti_tmp = beta_ti + step_ti;
          logplkd_list = logplkd_fixtra_bresties(Z_tv, B1, B2, theta_tmp, Z_ti,
                                                 beta_ti_tmp, ti, n_strata, idx_B_sp,
                                                 idx_fail, n_Z_strata, idx_Z_strata,
                                                 istart, iend, parallel, threads);
          double objfun_tmp = logplkd_list["logplkd"];
          diff_objfun = objfun_tmp - objfun;
        }
        theta = theta_tmp; beta_ti = beta_ti_tmp;
        theta_list.push_back(theta); beta_ti_list.push_back(beta_ti);
        if (iter==1) objfun_init = objfun;
        if (stop=="incre")
          crit = inc / 2;
        else if (stop=="relch")
          crit = abs(diff_objfun/(diff_objfun+objfun));
        else if (stop=="ratch")
          crit = abs(diff_objfun/(diff_objfun+objfun-objfun_init));
        objfun += diff_objfun;
      }
    } else {
      mat step = update_list["step"];
      inc = update_list["inc"];
      if (btr=="none") {
        theta += step;
        theta_list.push_back(theta);
        crit = inc / 2;
        logplkd_list = logplkd_fixtra_bresties(Z_tv, B1, B2, theta, Z_ti, beta_ti,
                                               ti, n_strata, idx_B_sp, idx_fail,
                                               n_Z_strata, idx_Z_strata,
                                               istart, iend, parallel, threads);
        objfun = logplkd_list["logplkd"];
      } else {
        mat theta_tmp = theta + step;
        logplkd_list = logplkd_fixtra_bresties(Z_tv, B1, B2, theta_tmp, Z_ti,
                                               beta_ti, ti, n_strata, idx_B_sp,
                                               idx_fail, n_Z_strata, idx_Z_strata,
                                               istart, iend, parallel, threads);
        double objfun_tmp = logplkd_list["logplkd"];
        diff_objfun = objfun_tmp - objfun;
        if (btr=="dynamic")      rhs_btr = inc;
        else if (btr=="static")  rhs_btr = 1.0;
        while (diff_objfun < s * v * rhs_btr && btr_ct < btr_max) {
          ++btr_ct;
          v *= t;
          theta_tmp = theta + v * step;
          logplkd_list = logplkd_fixtra_bresties(Z_tv, B1, B2, theta_tmp, Z_ti,
                                                 beta_ti, ti, n_strata, idx_B_sp,
                                                 idx_fail, n_Z_strata, idx_Z_strata,
                                                 istart, iend, parallel, threads);
          double objfun_tmp = logplkd_list["logplkd"];
          diff_objfun = objfun_tmp - objfun;
        }
        theta = theta_tmp;
        theta_list.push_back(theta);
        if (iter==1) objfun_init = objfun;
        if (stop=="incre")
          crit = inc / 2;
        else if (stop=="relch")
          crit = abs(diff_objfun/(diff_objfun+objfun));
        else if (stop=="ratch")
          crit = abs(diff_objfun/(diff_objfun+objfun-objfun_init));
        objfun += diff_objfun;
      }
    }
    cout << "Iter " << iter << ": Obj fun = " << setprecision(7) << fixed <<
      objfun << "; Stopping crit = " << setprecision(7) << scientific <<
        crit << ";" << endl;
    objfun_vec.push_back(objfun);
  }
  cout << "Algorithm converged after " << iter << " iterations!" << endl;
  CharacterVector names_strata = count_strata.names();
  vector<vec> hazard = logplkd_list["hazard"];
  List hazard_list;
  for (unsigned int i = 0; i < n_strata; ++i) {
    hazard_list.push_back(hazard[i], as<string>(names_strata[i]));
  }
  if (ti) {
    return List::create(_["ctrl.pts"]=theta, _["ctrl.pts.hist"]=theta_list,
                         _["tief"]=beta_ti,
                         _["tief.hist"]=beta_ti_list,
                         _["info"]=update_list["info"],
                         _["objfun"]=objfun*event.n_elem,
                         _["objfun.hist"]=objfun_vec*event.n_elem,
                         _["btr.ct"]=btr_ct,
                         _["hazard"]=hazard_list);
  } else {
    return List::create(_["ctrl.pts"]=theta, _["ctrl.pts.hist"]=theta_list,
                         _["info"]=update_list["info"],
                         _["objfun"]=objfun*event.n_elem,
                         _["objfun.hist"]=objfun_vec*event.n_elem,
                         _["btr.ct"]=btr_ct,
                         _["hazard"]=hazard_list);
  }
}

// [[Rcpp::export]]
List surtiver_fixtra_bresties_fit_diffpen(const vec &event, const vec &time, 
                                          const IntegerVector &count_strata,
                                          const mat &Z_tv, const mat &B1,
                                          const mat &B2,
                                          const mat &theta_init,
                                          const mat &Z_ti, const vec &beta_ti_init,
                                          const mat &D1, const mat &D2,
                                          const vec &mu1, const vec &mu2,
                                          const string &method="Newton",
                                          const double lambda=1e8, 
                                          const double &factor=1.0,
                                          const bool &parallel=false, 
                                          const unsigned int &threads=1,
                                          const double &tol=1e-10,
                                          const unsigned int &iter_max=20,
                                          const double &s=1e-2, const double &t=0.6,
                                          const string &btr="dynamic",
                                          const string &stop="incre") {
  // B1 has the same number of rows as Z_tv and Z_ti
  // B2 only includes distinct failure times sorted by strata
  bool ti = norm(Z_ti, "inf") > sqrt(datum::eps);
  IntegerVector cumsum_strata = cumsum(count_strata);
  const unsigned int n_strata = cumsum_strata.length();
  cumsum_strata.push_front(0);
  vector<vector<uvec>> idx_fail;
  vector<vector<unsigned int>> idx_fail_1st, idx_Z_strata;
  vector<unsigned int> n_fail_time, n_Z_strata;
  // each element of idx_Z_strata contains start/end row indices of Z_strata
  for (unsigned int i = 0; i < n_strata; ++i) {
    vec event_tmp = event.rows(cumsum_strata[i], cumsum_strata[i+1]-1),
      time_tmp = time.rows(cumsum_strata[i], cumsum_strata[i+1]-1);
    uvec idx_fail_tmp = find(event_tmp==1);
    vector<unsigned int> idx_Z_strata_tmp;
    idx_Z_strata_tmp.push_back(cumsum_strata[i]+idx_fail_tmp(0));
    idx_Z_strata_tmp.push_back(cumsum_strata[i+1]-1);
    idx_Z_strata.push_back(idx_Z_strata_tmp);
    n_Z_strata.push_back(cumsum_strata[i+1]-cumsum_strata[i]-
      idx_fail_tmp(0));
    vec time_fail_tmp = time_tmp.elem(idx_fail_tmp);
    idx_fail_tmp -= idx_fail_tmp(0);
    vec uniq_t = unique(time_fail_tmp);
    n_fail_time.push_back(uniq_t.n_elem);
    vector<uvec> idx_fail_tmp_tmp;
    vector<unsigned int> idx_fail_1st_tmp;
    for (vec::iterator j = uniq_t.begin(); j < uniq_t.end(); ++j) {
      uvec tmp = idx_fail_tmp.elem(find(time_fail_tmp==*j));
      idx_fail_tmp_tmp.push_back(tmp);
      idx_fail_1st_tmp.push_back(tmp[0]);
    }
    idx_fail.push_back(idx_fail_tmp_tmp);
    idx_fail_1st.push_back(idx_fail_1st_tmp);
  }
  IntegerVector n_failtime = wrap(n_fail_time);
  IntegerVector cumsum_failtime = cumsum(n_failtime);
  cumsum_failtime.push_front(0);
  vector<uvec> idx_B_sp;
  for (unsigned int i = 0; i < n_strata; ++i) {
    idx_B_sp.push_back(regspace<uvec>(cumsum_failtime[i],
                                      cumsum_failtime[i+1]-1));
  }
  // istart and iend for each thread when parallel=true
  vector<vec> cumsum_ar;
  vector<vector<unsigned int>> istart, iend;
  if (parallel) {
    for (unsigned int i = 0; i < n_strata; ++i) {
      double scale_fac = as_scalar(idx_fail[i].back().tail(1));
      cumsum_ar.push_back(
        (double)n_Z_strata[i] / scale_fac * regspace(1, idx_fail[i].size()) -
          cumsum(conv_to<vec>::from(idx_fail_1st[i])/scale_fac));
      vector<unsigned int> istart_tmp, iend_tmp;
      for (unsigned int id = 0; id < threads; ++id) {
        istart_tmp.push_back(as_scalar(find(cumsum_ar[i] >=
          cumsum_ar[i](idx_fail[i].size()-1)/(double)threads*id, 1)));
        iend_tmp.push_back(as_scalar(find(cumsum_ar[i] >=
          cumsum_ar[i](idx_fail[i].size()-1)/(double)threads*(id+1), 1)));
        if (id == threads-1) {
          iend_tmp.pop_back();
          iend_tmp.push_back(idx_fail[i].size());
        }
      }
      istart.push_back(istart_tmp);
      iend.push_back(iend_tmp);
    }
  }
  // iterative algorithm with backtracking line search
  unsigned int iter = 0, btr_max = 100, btr_ct = 0;
  mat theta = theta_init; vec beta_ti = beta_ti_init;
  List theta_list = List::create(theta), beta_ti_list = List::create(beta_ti);
  double crit = 1.0, v = 1.0, objfun_init = 0, objfun, diff_objfun,
    inc, rhs_btr = 0;
  List logplkd_list, update_list;
  NumericVector objfun_vec;
  logplkd_list =
    logplkd_fixtra_bresties_diffpen(Z_tv, B1, B2, theta, Z_ti, beta_ti,
                                    ti, n_strata, D1, D2, mu1, mu2,
                                    idx_B_sp, idx_fail,
                                    n_Z_strata, idx_Z_strata,
                                    istart, iend, parallel, threads);
  objfun = logplkd_list["logplkd"];
  objfun_vec.push_back(objfun);
  while (iter < iter_max && crit > tol) {
    ++iter;
    // if (iter > 1) break;
    update_list =
    stepinc_fixtra_bresties_diffpen(Z_tv, B1, B2, theta, Z_ti, beta_ti,
                                    ti, n_strata, D1, D2, mu1, mu2,
                                    idx_B_sp, idx_fail,
                                    idx_Z_strata, istart, iend, method,
                                    lambda*pow(factor,iter-1),
                                    parallel, threads);
    v = 1.0; // reset step size
    if (ti) {
      mat step_tv = update_list["step_tv"];
      vec step_ti = update_list["step_ti"];
      inc = update_list["inc"];
      // if (iter==1) {
      //   Rcout << inc << endl;
      // }
      if (btr=="none") {
        theta += step_tv;
        beta_ti += step_ti;
        theta_list.push_back(theta); beta_ti_list.push_back(beta_ti);
        crit = inc / 2;
        logplkd_list =
          logplkd_fixtra_bresties_diffpen(Z_tv, B1, B2, theta, Z_ti, beta_ti,
                                          ti, n_strata, D1, D2, mu1, mu2,
                                          idx_B_sp, idx_fail,
                                          n_Z_strata, idx_Z_strata,
                                          istart, iend, parallel, threads);
        objfun = logplkd_list["logplkd"];
      } else {
        mat theta_tmp = theta + step_tv;
        vec beta_ti_tmp = beta_ti + step_ti;
        logplkd_list =
          logplkd_fixtra_bresties_diffpen(Z_tv, B1, B2, theta_tmp, Z_ti,
                                          beta_ti_tmp, ti, n_strata,
                                          D1, D2, mu1, mu2,
                                          idx_B_sp, idx_fail,
                                          n_Z_strata, idx_Z_strata,
                                          istart, iend, parallel, threads);
        double objfun_tmp = logplkd_list["logplkd"];
        diff_objfun = objfun_tmp - objfun;
        if (btr=="dynamic")      rhs_btr = inc;
        else if (btr=="static")  rhs_btr = 1.0;
        while (diff_objfun < s * v * rhs_btr && btr_ct < btr_max) {
          ++btr_ct;
          v *= t;
          theta_tmp = theta + v * step_tv;
          beta_ti_tmp = beta_ti + step_ti;
          logplkd_list =
            logplkd_fixtra_bresties_diffpen(Z_tv, B1, B2, theta_tmp, Z_ti,
                                            beta_ti_tmp, ti, n_strata,
                                            D1, D2, mu1, mu2, idx_B_sp,
                                            idx_fail, n_Z_strata, idx_Z_strata,
                                            istart, iend, parallel, threads);
          double objfun_tmp = logplkd_list["logplkd"];
          diff_objfun = objfun_tmp - objfun;
        }
        theta = theta_tmp; beta_ti = beta_ti_tmp;
        theta_list.push_back(theta); beta_ti_list.push_back(beta_ti);
        if (iter==1) objfun_init = objfun;
        if (stop=="incre")
          crit = inc / 2;
        else if (stop=="relch")
          crit = abs(diff_objfun/(diff_objfun+objfun));
        else if (stop=="ratch")
          crit = abs(diff_objfun/(diff_objfun+objfun-objfun_init));
        objfun += diff_objfun;
      }
    } else {
      mat step = update_list["step"];
      inc = update_list["inc"];
      if (btr=="none") {
        theta += step;
        theta_list.push_back(theta);
        crit = inc / 2;
        logplkd_list =
          logplkd_fixtra_bresties_diffpen(Z_tv, B1, B2, theta, Z_ti, beta_ti,
                                          ti, n_strata, D1, D2, mu1, mu2,
                                          idx_B_sp, idx_fail,
                                          n_Z_strata, idx_Z_strata,
                                          istart, iend, parallel, threads);
        objfun = logplkd_list["logplkd"];
      } else {
        mat theta_tmp = theta + step;
        logplkd_list =
          logplkd_fixtra_bresties_diffpen(Z_tv, B1, B2, theta_tmp, Z_ti,
                                          beta_ti, ti, n_strata, D1, D2, mu1, mu2,
                                          idx_B_sp,
                                          idx_fail, n_Z_strata, idx_Z_strata,
                                          istart, iend, parallel, threads);
        double objfun_tmp = logplkd_list["logplkd"];
        diff_objfun = objfun_tmp - objfun;
        if (btr=="dynamic")      rhs_btr = inc;
        else if (btr=="static")  rhs_btr = 1.0;
        while (diff_objfun < s * v * rhs_btr && btr_ct < btr_max) {
          ++btr_ct;
          v *= t;
          theta_tmp = theta + v * step;
          logplkd_list =
            logplkd_fixtra_bresties_diffpen(Z_tv, B1, B2, theta_tmp, Z_ti,
                                            beta_ti, ti, n_strata,
                                            D1, D2, mu1, mu2, idx_B_sp,
                                            idx_fail, n_Z_strata, idx_Z_strata,
                                            istart, iend, parallel, threads);
          double objfun_tmp = logplkd_list["logplkd"];
          diff_objfun = objfun_tmp - objfun;
        }
        theta = theta_tmp;
        theta_list.push_back(theta);
        if (iter==1) objfun_init = objfun;
        if (stop=="incre")
          crit = inc / 2;
        else if (stop=="relch")
          crit = abs(diff_objfun/(diff_objfun+objfun));
        else if (stop=="ratch")
          crit = abs(diff_objfun/(diff_objfun+objfun-objfun_init));
        objfun += diff_objfun;
      }
    }
    cout << "Iter " << iter << ": Obj fun = " << setprecision(7) << fixed <<
      objfun << "; Stopping crit = " << setprecision(7) << scientific <<
        crit << ";" << endl;
    objfun_vec.push_back(objfun);
  }
  cout << "Algorithm converged after " << iter << " iterations!" << endl;
  CharacterVector names_strata = count_strata.names();
  vector<vec> hazard = logplkd_list["hazard"];
  List hazard_list;
  for (unsigned int i = 0; i < n_strata; ++i) {
    hazard_list.push_back(hazard[i], as<string>(names_strata[i]));
  }
  if (ti) {
    return List::create(_["ctrl.pts"]=theta, _["ctrl.pts.hist"]=theta_list,
                        _["tief"]=beta_ti,
                        _["tief.hist"]=beta_ti_list,
                        _["info"]=update_list["info"],
                        _["objfun"]=objfun*event.n_elem,
                        _["objfun.hist"]=objfun_vec*event.n_elem,
                        _["btr.ct"]=btr_ct,
                        _["hazard"]=hazard_list);
  } else {
    return List::create(_["ctrl.pts"]=theta, _["ctrl.pts.hist"]=theta_list,
                        _["info"]=update_list["info"],
                        _["objfun"]=objfun*event.n_elem,
                        _["objfun.hist"]=objfun_vec*event.n_elem,
                        _["btr.ct"]=btr_ct,
                        _["hazard"]=hazard_list);
  }
}